import asyncio
import functools
import inspect
import logging
import time
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    NoReturn,
    TypeVar,
    Union,
    overload,
)

from a_sync import ASyncGenericBase
from a_sync.a_sync.method import ASyncBoundMethod
from a_sync.iter import ASyncGeneratorFunction
from typing_extensions import Concatenate, ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


def continue_on_revert(func: Callable[P, T]) -> Callable[P, T]:
    """
    Decorates a call-making function. If the call reverts, returns `None`. Raises any other exceptions.
    """
    from y.exceptions import continue_if_call_reverted

    if asyncio.iscoroutinefunction(func):

        @functools.wraps(func)
        async def continue_on_revert_wrap(
            *args: P.args, **kwargs: P.kwargs
        ) -> Union[T, None]:
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                continue_if_call_reverted(e)

    elif callable(func):

        @functools.wraps(func)
        def continue_on_revert_wrap(
            *args: P.args, **kwargs: P.kwargs
        ) -> Union[T, None]:
            try:
                return func(*args, **kwargs)
            except Exception as e:
                continue_if_call_reverted(e)

    else:
        raise NotImplementedError(f"Unable to decorate {func}")
    return continue_on_revert_wrap


B = TypeVar("B", bound=ASyncGenericBase)


@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[B, P], AsyncIterator[T]]
) -> ASyncGeneratorFunction[P, T]: ...


@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[B, P], Awaitable[T]]
) -> ASyncBoundMethod[B, P, T]: ...


@overload
def stuck_coro_debugger(
    fn: Callable[Concatenate[B, P], T]
) -> ASyncBoundMethod[B, P, T]: ...


@overload
def stuck_coro_debugger(
    fn: Callable[P, AsyncIterator[T]]
) -> Callable[P, AsyncIterator[T]]: ...


@overload
def stuck_coro_debugger(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]: ...


def stuck_coro_debugger(fn):
    logger = logging.getLogger("y.stuck?")
    if inspect.isasyncgenfunction(fn):

        @functools.wraps(fn)
        async def stuck_async_gen_wrap(
            *args: P.args, **kwargs: P.kwargs
        ) -> AsyncIterator[T]:
            if not logger.isEnabledFor(logging.DEBUG):
                async for thing in fn(*args, **kwargs):
                    yield thing
            t = asyncio.create_task(
                coro=_stuck_debug_task(logger, fn, args, kwargs),
                name="_stuck_debug_task",
            )
            try:
                async for thing in fn(*args, **kwargs):
                    yield thing
                t.cancel()
                return
            except Exception as e:
                t.cancel()
                raise

        return stuck_async_gen_wrap
    else:

        @functools.wraps(fn)
        async def stuck_coro_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            if not logger.isEnabledFor(logging.DEBUG):
                return await fn(*args, **kwargs)
            t = asyncio.create_task(
                coro=_stuck_debug_task(logger, fn, args, kwargs),
                name="_stuck_debug_task",
            )
            try:
                retval = await fn(*args, **kwargs)
                t.cancel()
            except Exception as e:
                t.cancel()
                raise
            return retval

        return stuck_coro_wrap


async def _stuck_debug_task(
    logger: logging.Logger, fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs
) -> NoReturn:
    start = time.time()
    while True:
        await asyncio.sleep(300)
        logger._log(
            logging.DEBUG,
            f"{fn.__module__}.{fn.__name__} still executing after {round((time.time() - start)/60, 2)}m with"
            + f" args {tuple(str(arg) for arg in args)}"
            + f" kwargs {dict((k, str(v)) for k, v in kwargs.items())}",
            (),
        )



from typing import Optional

import a_sync
from brownie import chain
from eth_typing import ChecksumAddress

from y import ENVIRONMENT_VARIABLES as ENVS
from y.datatypes import Block, UsdPrice
from y.networks import Network
from y.prices import magic

MAPPING = {
    Network.Mainnet: {
        "0x4da27a545c0c5B758a6BA100e3a049001de870f5": "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9",  # stkaave -> aave
        "0x27D22A7648e955E510a40bDb058333E9190d12D4": "0x0cec1a9154ff802e7934fc916ed7ca50bde6844e",  # ppool -> pool
        # TODO: algorithmically get gauges
        "0xcF5136C67fA8A375BaBbDf13c0307EF994b5681D": "0x425BfB93370F14fF525aDb6EaEAcfE1f4e3b5802",  # sdai-usdm-gauge -> sdai-usdm
        "0x590f7e2b211Fa5Ff7840Dd3c425B543363797701": "0x5756bbdDC03DaB01a3900F01Fb15641C3bfcc457",  # YFImkUSD-gauge -> YFImkUSD
    },
}.get(chain.id, {})


def is_one_to_one_token(token_address: ChecksumAddress) -> bool:
    return token_address in MAPPING


@a_sync.a_sync(default="sync")
async def get_price(
    token_address: ChecksumAddress,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    return await magic.get_price(
        MAPPING[token_address], block=block, skip_cache=skip_cache, sync=False
    )



import logging
import threading
from typing import Dict, Optional

import a_sync
from brownie import chain
from cachetools import TTLCache, cached
from pony.orm import select
from y._db.decorators import (
    a_sync_read_db_session,
    a_sync_write_db_session,
    log_result_count,
)
from y._db.entities import Contract
from y._db.utils._ep import _get_get_token
from y._db.utils.utils import ensure_block
from y.datatypes import Address, Block


logger = logging.getLogger(__name__)

_get_get_token()


@a_sync_read_db_session
def get_deploy_block(address: str) -> Optional[int]:
    if deploy_block := known_deploy_blocks().pop(address, None):
        logger.debug("%s deploy block from cache: %s", address, deploy_block)
        return deploy_block
    get_token = _get_get_token()
    if deploy_block := get_token(address, sync=True).deploy_block:
        logger.debug("%s deploy block from cache: %s", address, deploy_block.number)
        return deploy_block.number
    logger.debug("%s deploy block not cached, fetching from chain", address)


async def _set_deploy_block(address: str, deploy_block: int) -> None:
    await ensure_block(deploy_block)
    return await __set_deploy_block(address, deploy_block)


set_deploy_block = a_sync.ProcessingQueue(
    _set_deploy_block, num_workers=10, return_data=False
)


@a_sync_write_db_session
def __set_deploy_block(address: str, deploy_block: int) -> None:
    from y._db.utils._ep import _get_get_token

    get_token = _get_get_token()
    get_token(address, sync=True).deploy_block = (chain.id, deploy_block)
    logger.debug("deploy block cached for %s: %s", address, deploy_block)


# startup caches


@cached(TTLCache(maxsize=1, ttl=60 * 60), lock=threading.Lock())
@log_result_count("deploy blocks")
def known_deploy_blocks() -> Dict[Address, Block]:
    """cache and return all known contract deploy blocks for this chain to minimize db reads"""
    return dict(
        select(
            (c.address, c.deploy_block.number)
            for c in Contract
            if c.chain.id == chain.id and c.deploy_block.number
        )
    )



import asyncio
from typing import List, Optional, Tuple

from y._decorators import continue_on_revert, stuck_coro_debugger
from y.datatypes import Address, Block
from y.exceptions import call_reverted
from y.prices.dex.uniswap.v2 import Path, UniswapRouterV2, UniswapV2Pool
from y.utils.cache import a_sync_ttl_cache


class SolidlyRouterBase(UniswapRouterV2):
    """
    Solidly is a modified fork of Uni V2.
    The `uniswap_multiplexer` is the entrypoint for pricing using this object.
    """

    @continue_on_revert
    @stuck_coro_debugger
    async def get_quote(
        self, amount_in: int, path: Path, block: Optional[Block] = None
    ) -> Tuple[int, int]:
        routes = await self.get_routes_from_path(path, block)
        try:
            return await self.contract.getAmountsOut.coroutine(
                amount_in, routes, block_identifier=block
            )
        except Exception as e:
            strings = [
                "INSUFFICIENT_INPUT_AMOUNT",
                "INSUFFICIENT_LIQUIDITY",
                "INSUFFICIENT_OUT_LIQUIDITY",
                "Sequence has incorrect length",
            ]
            if not call_reverted(e) and all(s not in str(e) for s in strings):
                raise


class SolidlyPool(UniswapV2Pool):
    pass


class SolidlyRouter(SolidlyRouterBase):
    @stuck_coro_debugger
    @a_sync_ttl_cache
    async def pair_for(
        self, input_token: Address, output_token: Address, stable: bool
    ) -> Address:
        return await self.contract.pairFor.coroutine(input_token, output_token, stable)

    @stuck_coro_debugger
    @a_sync_ttl_cache
    async def get_pool(
        self, input_token: Address, output_token: Address, stable: bool, block: Block
    ) -> Optional[SolidlyPool]:
        pool_address = await self.pair_for(
            input_token, output_token, stable, sync=False
        )
        if await self.contract.isPair.coroutine(pool_address, block_identifier=block):
            return SolidlyPool(pool_address, asynchronous=self.asynchronous)

    @stuck_coro_debugger
    async def get_routes_from_path(
        self, path: Path, block: Block
    ) -> List[Tuple[Address, Address, bool]]:
        routes = []
        for i in range(len(path) - 1):
            input_token, output_token = path[i], path[i + 1]
            # Try for a stable pool first and use that if available
            stable_pool, unstable_pool = await asyncio.gather(
                self.get_pool(input_token, output_token, True, block, sync=False),
                self.get_pool(input_token, output_token, False, block, sync=False),
            )

            if stable_pool and unstable_pool:
                # We have to find out which of these pools is deepest
                stable_reserves, unstable_reserves = await asyncio.gather(
                    stable_pool.reserves(block=block, sync=False),
                    unstable_pool.reserves(block=block, sync=False),
                )
                stable_reserves = tuple(stable_reserves)
                unstable_reserves = tuple(unstable_reserves)

                # NOTE: using `__token0__` and `__token1__` is faster than `__tokens__` since they're already cached and return instantly
                #       it also creates 2 fewer tasks and 1 fewer future than `__tokens__` since there is no use of `asyncio.gather`.
                if (
                    await stable_pool.__token0__ == await unstable_pool.__token0__
                    and await stable_pool.__token1__ == await unstable_pool.__token1__
                ):
                    stable_reserve = stable_reserves[0]
                    unstable_reserve = unstable_reserves[0]
                else:  # Order of tokens is flip flopped in the pools
                    stable_reserve = stable_reserves[0]
                    unstable_reserve = unstable_reserves[1]
                if stable_reserve >= unstable_reserve:
                    is_stable = True
                elif stable_reserve < unstable_reserve:
                    is_stable = False
                routes.append([input_token, output_token, is_stable])
            elif stable_pool:
                routes.append([input_token, output_token, True])
            elif unstable_pool:
                routes.append([input_token, output_token, False])
            else:
                raise ValueError(
                    "Not sure why this function is even running if no pool is found"
                )

        return routes



from random import randint

from brownie.convert.datatypes import EthAddress
from y import convert
from y.classes.common import ContractBase
from y.constants import WRAPPED_GAS_COIN
from y.contracts import Contract


def test_contract_singleton():
    token = WRAPPED_GAS_COIN
    variations = [
        token.lower(),
        token.upper(),
        convert.to_address(token),
        Contract(token),
        EthAddress(token.lower()),
    ]
    for variation in variations:
        assert ContractBase(WRAPPED_GAS_COIN) == ContractBase(
            variation
        ), "ContractSingleton metaclass is not working."
